@@ -14,7 +14,7 @@ from motorcortex.setup_logger import logger
 from motorcortex.state_callback_handler import StateCallbackHandler
 
 import os
-from threading import Condition  # Add this import
+from threading import Event  # Add this import
 from concurrent.futures import ThreadPoolExecutor
 from pynng import Req0, TLSConfig
 from enum import Enum
@@ -29,98 +29,120 @@ class ConnectionState(Enum):
     DISCONNECTED = 5
 
 
-class Base64Encoder(json.JSONEncoder):
-    # pylint: disable=method-hidden
-    encoding = 'utf-8'
-
-    def default(self, o):
-        if isinstance(o, bytes):
-            return base64.b64encode(o).decode(Base64Encoder.encoding)
-        return json.JSONEncoder.default(self, o)
-
-
-class Base64Decoder(json.JSONDecoder):
-    # pylint: disable=method-hidden
-    encoding = 'utf-8'
-
-    def default(self, o):
-        if isinstance(o, bytes):
-            return base64.b64decode(o).encode(Base64Decoder.decode())
-        return json.Base64Decoder.default(self, o)
-
-
 class Request(object):
 
     def __init__(self, protobuf_types, parameter_tree):
         self.__socket = None
         self.__url = None
+        self.__connected_event = None
         self.__connected = False
-        self.__connected_lock = None
         self.__protobuf_types = protobuf_types
         self.__parameter_tree = parameter_tree
         self.__connection_state = ConnectionState.DISCONNECTED
         self.__pool = ThreadPoolExecutor(max_workers=1)
         self.__callback_handler = StateCallbackHandler()
+        logger.debug("Request object initialized with state: DISCONNECTED")
 
     def url(self):
         return self.__url
 
     def connect(self, url, **kwargs):
+        logger.debug(f"[CONNECT] Starting connection to {url}")
+        logger.debug(f"[CONNECT] Current state: {self.__connection_state.name}")
+
         self.__connection_state = ConnectionState.CONNECTING
         conn_timeout_ms, recv_timeout_ms, certificate, state_update = self.parse(**kwargs)
 
+        logger.debug(f"[CONNECT] Parameters - timeout: {conn_timeout_ms} ms, recv_timeout: {recv_timeout_ms} ms, "
+                     f"has_cert: {certificate is not None}, has_state_update: {state_update is not None}")
+
         if state_update:
+            logger.debug("[CONNECT] Starting state callback handler")
             self.__callback_handler.start(state_update)
 
         self.__url = url
         tls_config = None
         if certificate:
+            logger.debug(f"[CONNECT] Using TLS with certificate: {certificate}")
             tls_config = TLSConfig(TLSConfig.MODE_CLIENT, ca_files=certificate)
 
+        logger.debug(f"[CONNECT] Creating Req0 socket with recv_timeout={recv_timeout_ms}ms")
         self.__socket = Req0(recv_timeout=recv_timeout_ms, tls_config=tls_config)
-        self.__connected_lock = Condition()  # Create Condition instead of Queue
-        self.__connected = False  # Initialize state
+        self.__connected_event = Event()  # Create a fresh event for each connection
+        self.__connected = False  # Reset state
 
-        def pre_connect_cb(pipe):
-            with self.__connected_lock:
-                self.__connected = True
-                self.__connection_state = ConnectionState.CONNECTION_OK
-                self.__callback_handler.notify(self, self.connectionState())
-                self.__connected_lock.notify_all()  # Wake up all waiting threads
+        logger.debug("[CONNECT] Socket created, registering callbacks")
 
-        def post_remove_cb(pipe):
-            with self.__connected_lock:
-                if self.__connection_state == ConnectionState.DISCONNECTING:
-                    self.__connection_state = ConnectionState.DISCONNECTED
-                elif self.__connection_state == ConnectionState.CONNECTING:
-                    self.__connection_state = ConnectionState.CONNECTION_FAILED
-                elif self.__connection_state == ConnectionState.CONNECTION_OK:
-                    self.__connection_state = ConnectionState.CONNECTION_LOST
-                self.__connected = False
-                self.__callback_handler.notify(self, self.connectionState())
-                self.__connected_lock.notify_all()  # Wake up all waiting threads
+        def pre_connect_cb(_pipe):
+            logger.debug(f"[CALLBACK] PRE_CONNECT fired - Connection established")
+            old_state = self.__connection_state.name
+            self.__connected = True
+            self.__connection_state = ConnectionState.CONNECTION_OK
+            logger.debug(f"[CALLBACK] State transition: {old_state} -> {self.__connection_state.name}")
+            self.__callback_handler.notify(self, self.connectionState())
+            self.__connected_event.set()
+            logger.debug("[CALLBACK] Connection event set, connected=True")
 
-        self.__socket.add_post_pipe_connect_cb(pre_connect_cb)
+        def post_remove_cb(_pipe):
+            logger.debug(f"[CALLBACK] POST_REMOVE fired - Connection lost/failed")
+            old_state = self.__connection_state.name
+
+            if self.__connection_state == ConnectionState.DISCONNECTING:
+                self.__connection_state = ConnectionState.DISCONNECTED
+                logger.debug(f"[CALLBACK] Clean disconnect: {old_state} -> DISCONNECTED")
+            elif self.__connection_state == ConnectionState.CONNECTING:
+                self.__connection_state = ConnectionState.CONNECTION_FAILED
+                logger.debug(f"[CALLBACK] Connection failed: {old_state} -> CONNECTION_FAILED")
+            elif self.__connection_state == ConnectionState.CONNECTION_OK:
+                self.__connection_state = ConnectionState.CONNECTION_LOST
+                logger.debug(f"[CALLBACK] Connection lost: {old_state} -> CONNECTION_LOST")
+
+            self.__connected = False
+            self.__callback_handler.notify(self, self.connectionState())
+            self.__connected_event.set()
+            logger.debug("[CALLBACK] Connection event set, connected=False")
+
+        self.__socket.add_pre_pipe_connect_cb(pre_connect_cb)
         self.__socket.add_post_pipe_remove_cb(post_remove_cb)
+        logger.debug("[CONNECT] Callbacks registered successfully")
+
+        logger.debug(f"[CONNECT] Starting dial to {url} (non-blocking)")
         self.__socket.dial(url, block=False)
 
-        return Reply(self.__pool.submit(self.waitForConnection, self.__connected_lock,
-                                        conn_timeout_ms / 1000.0))
+        logger.debug(f"[CONNECT] Submitting waitForConnection to thread pool with timeout={conn_timeout_ms / 1000.0}s")
+        return Reply(self.__pool.submit(self.waitForConnection, self.__connected_event,
+                                        conn_timeout_ms / 1000.0, lambda: self.__connected))
 
     def close(self):
+        logger.debug("[CLOSE] Closing connection")
+        logger.debug(f"[CLOSE] Current state: {self.__connection_state.name}, connected: {self.__connected}")
+
         self.__connection_state = ConnectionState.DISCONNECTING
-        if self.__connected_lock:
-            with self.__connected_lock:
-                self.__connected = False
-                self.__connected_lock.notify_all()
-        self.__socket.close()
+        if self.__connected_event:
+            self.__connected = False
+            self.__connected_event.set()
+            logger.debug("[CLOSE] Connection event set for shutdown")
+        else:
+            logger.debug("[CLOSE] No connection event to set")
+
+        if self.__socket:
+            logger.debug("[CLOSE] Closing socket")
+            self.__socket.close()
+        else:
+            logger.debug("[CLOSE] No socket to close")
+
+        logger.debug("[CLOSE] Stopping callback handler")
         self.__callback_handler.stop()
+
+        logger.debug("[CLOSE] Shutting down thread pool (blocking)")
         self.__pool.shutdown(wait=True)
+        logger.debug("[CLOSE] Connection closed successfully")
 
     def send(self, encoded_msg, do_not_decode_reply=False):
         if self.__socket is not None:
             return Reply(self.__pool.submit(self.__send, self.__socket, encoded_msg,
                                             None if do_not_decode_reply else self.__protobuf_types))
+        logger.debug("[SEND] Attempted to send on null socket - connection not established?")
         return None
 
     def login(self, login, password):
@@ -155,7 +177,7 @@ class Request(object):
         """Request a parameter tree hash from the server.
 
             Returns:
-                Reply(ParameterTreeMsg): A Promise, which resolves when parameter tree hash is received or fails
+                Reply(ParameterTreeMsg): A Promise, which resolves when a parameter tree hash is received or fails
                 otherwise. ParameterTreeHashMsg message has a status field to check the status of the operation.
 
             Examples:
@@ -188,14 +210,14 @@ class Request(object):
                                         self.getParameterTreeHash(), self.__protobuf_types, self.__socket))
 
     def save(self, path, file_name):
-        """Request a server to save a parameter tree to file.
+        """Request a server to save a parameter tree to a file.
 
             Args:
                 path(str): path where to save
                 file_name(str): file name
 
             Returns:
-                Reply(StatusMsg): A promise, which resolves when save operation is completed,
+                Reply(StatusMsg): A promise, which resolves when the save operation is completed,
                 fails otherwise.
 
         """
@@ -218,7 +240,7 @@ class Request(object):
                 value argument)
 
             Returns:
-                  Reply(StatusMsg): A Promise, which resolves when parameter value is updated or fails otherwise.
+                  Reply(StatusMsg): A Promise, which resolves when a parameter value is updated or fails otherwise.
 
             Examples:
                   >>> reply = req.setParameter("root/Control/activateSemiAuto", False)
@@ -497,28 +519,50 @@ class Request(object):
 
     @staticmethod
     def __send(req, encoded_msg, protobuf_types):
-        req.send(encoded_msg)
-        buffer = req.recv()
-        if buffer:
-            if protobuf_types:
-                return protobuf_types.decode(buffer)
-            else:
-                return buffer
+        try:
+            req.send(encoded_msg)
+            buffer = req.recv()
+            if buffer:
+                if protobuf_types:
+                    return protobuf_types.decode(buffer)
+                else:
+                    return buffer
+        except Exception as e:
+            logger.error(f"[__SEND] Error during send/recv: {type(e).__name__}: {e}")
+            raise
 
         return None
 
     @staticmethod
-    def waitForConnection(condition, timeout_sec):
-        with condition:
-            # Wait for the condition to be set or timeout
-            if timeout_sec <= 0:
-                result = condition.wait()
-            else:
-                result = condition.wait(timeout_sec)
+    def waitForConnection(event, timeout_sec, is_connected_fn=None):
+        logger.debug(f"[WAIT] waitForConnection started with timeout={timeout_sec}s")
 
-            if not result:
-                raise TimeoutError("Connection timeout")
-            return True
+        # Wait for the condition to be set or timeout
+        if timeout_sec <= 0:
+            logger.debug("[WAIT] Waiting indefinitely for connection event")
+            result = event.wait()
+        else:
+            logger.debug(f"[WAIT] Waiting up to {timeout_sec}s for connection event")
+            result = event.wait(timeout_sec)
+
+        if not result:
+            logger.error(f"[WAIT] Connection timeout after {timeout_sec}s - no event received")
+            raise TimeoutError(f"Connection timeout after {timeout_sec}s")
+
+        logger.debug("[WAIT] Event received - checking connection status")
+
+        # Check if we actually connected (event could be set by close() or failure)
+        if is_connected_fn:
+            connected = is_connected_fn()
+            logger.debug(f"[WAIT] Connection status check: connected={connected}")
+            if not connected:
+                logger.error("[WAIT] Event was set but connection failed or was closed")
+                raise ConnectionError("Connection failed or was closed before completion")
+        else:
+            logger.debug("[WAIT] No connection status check function provided")
+
+        logger.debug("[WAIT] Connection successfully established")
+        return True
 
     @staticmethod
     def __getParameterTree(hash_reply, protobuf_types, socket):
